# -*- Mode: perl -*-
#--------------------------------------------------------------------------------
#  reference.xml => texi converter
#
#  HANAI Tohru <honey@din.or.jp>
#--------------------------------------------------------------------------------
# ひつようなもの。
#  ・perl 一式 ( http://www.activestate.com/ )
#  ・XML パッケージ(ActivePerl にならついてるはず。)
#  ・Jcode パッケージ( ftp://ftp.lab.kdd.co.jp/lang/perl/CPAN/authors/id/D/DA/DANKOGAI/ )
#
# つかいかた。
#  shell 上でこんな風に打ち込む：
#    $ perl ref2texi.pl
#  すると、.texi ファイル "$outfile" が ディレクトリ "$outdir"上にできるはず。
#  ($outxxx は、下のリスト中で定義されてるもの。)
#
# .info ファイルのつくりかた。
#  いまのところ xyzzy は .texi ファイルのことなんか知ったことじゃないみたい
#  なので、XEmacs とか meadow とかのコマンドを使うことに。
#   1. meadow か何かを立ち上げる。
#   2. .texi ファイル "$outfile" を読み込む。
#   3. 日本語がきちんと表示されていることを確認。されてなければなんとかする。
#   4. コンバート実行: C-c C-e C-b (M-x texinfo-format-buffer)
#   5. できあがった .info ファイルを保存: C-x C-s (M-x save-buffer)
#
# 以上。
# ---------- --- -
# ひとこと。
#  ・この perl スクリプトを使ってなにが起きても責任とれませんので。
#  ・再配布はご自由に。改変しての再配布もご自由に。
#                                                         1999年12月吉日 花井　達
#--------------------------------------------------------------------------------

use XML::DOM;
use Jcode;

$| = 1;

#------------------------------
# 処理
#------------------------------
sub main {
    ($outfile =~ /[ \t]/) and die "\$outfile に空白文字がはいっちゃってるので中止。($outfile)\n";
    
    parse_xml();        # xml を解析し、%sects を構築
    generate_texi();    # texi ファイル生成
    return;             # 以上。
}

#------------------------------
# 入出力ファイル
#------------------------------

# 出力文字セット (Jcode用)
local $out_charset = "sjis";
# 入力文字セット (Jcode用)
local $in_charset = "utf8";

# 入力ファイル名(XML)
local $datafile = "reference.xml";
# 出力ファイル名(texi)
local $outfile = "reference.texi";     # 空白文字($20)を含まないこと。(info の作成に失敗するから)
# 出力ディレクトリ
local $outdir = "./info";

#------------------------------
# 解析結果格納変数
#------------------------------

# セクション分類による全データハッシュ
local %sects = ();
# 目的別索引用
local %tips = ();

#------------------------------

#目次の順番リスト
local @INDEX_LIST = ("データ型",
                     "変数と定数",
                     "Predicates",
                     "制御構造",
                     "パッケージ",
                     "関数",
                     "シンボル",
                     "数値",
                     "文字",
                     "文字列",
                     "Sequence",
                     "リスト",
                     "ハッシュ",
                     "配列",
                     "チャンク",
                     "評価",
                     "入出力",
                     "ファイルシステム",
                     "エラー",
                     "ウィンドウ",
                     "バッファ",
                     "モード",
                     "キーマップ",
                     "テキスト",
                     "検索・正規表現",
                     "ダイアログ",
                     "日付・時間",
                     "メニュー",
                     "ファイラー",
                     "ポジション",
                     "プロセス",
                     "システム",
                     "その他",
                     "未分類"
                     );

# 目次の別名
local %INDEX_NAME_LIST = ("データ型" => "datatypes",
                          "変数と定数" => "variables",
                          "Predicates" => "predicates",
                          "制御構造" => "program",
                          "関数" => "function",
                          "シンボル" => "symbols",
                          "パッケージ" => "packages",
                          "数値" => "numbers",
                          "文字" => "characters",
                          "文字列" => "string",
                          "Sequence" => "sequence",
                          "リスト" => "list",
                          "ハッシュ" => "hash",
                          "配列" => "array", 
                          "チャンク" => "chunk",
                          "評価" => "evaluation",
                          "入出力" => "io",
                          "ファイルシステム" => "filesystem",
                          "エラー" => "errors",
                          "ウィンドウ" => "window",
                          "バッファ" => "buffer",
                          "モード" => "mode",
                          "キーマップ" => "keymap",
                          "メニュー" => "menu",
                          "日付・時間" => "time",
                          "ダイアログ" => "dialog",
                          "システム" => "system",
                          "ファイラー" => "filer",
                          "テキスト" => "text",
                          "プロセス" => "process",
                          "ポジション" => "position",
                          "検索・正規表現" => "search",
                          "その他" => "misc",
                          "未分類" => "other"
                          );

local %TEXI_CMD_REF = ('Function' => '@defun',
                       'Macro' => '@defmac',
                       'Misc' => 'misc: ',
                       'Special Form' => '@defspec',
                       'Tips' => 'tips: ',
                       'Variable' => '@defvar',
                       'Variables' => '@defvar'
                       );

sub parse_xml {
    my $parser = new XML::DOM::Parser;

    # XML Parse
    &report( sprintf( "parsing %s ... ", $datafile), 1);
    my $doc = $parser->parsefile( $datafile);
    &report( "done.\n");

    # データ処理
    &report( "studying data ... ", 1);
    # <book>
    for my $book ( $doc->getChildNodes){
        # <chapter>
        for my $chapter ( $book->getChildNodes){
            next unless ($chapter->getNodeType == ELEMENT_NODE);

            my $title;
            my %chap;
            # 各エレメント
            for my $element ( $chapter->getChildNodes){
                next unless ( $element->getNodeType == ELEMENT_NODE);

                my $name = $element->getNodeName;
                my $data;
                # エレメントの子供 (TEXT_NODEのはず)
                for my $key ( $element->getChildNodes){
                    if ( $key->getNodeType == TEXT_NODE){
                        $data = encode_text_for_target($key->getData);
                        if( $chap{ $name}){
                            $chap{ $name} .= "\t" . $data;
                        } else {
                            $chap{ $name} = $data;
                        }
                    }
                }
            }#for my $element

            # パースした chapter(1ヶ)を該当セクションにぶち込むのです。
            push( @ { $sects{ $chap{ "section"}}}, \%chap) if %chap;
            # 必要とあらば目的別索引用のハッシュにも。
            push( @ { $tips { $chap{ "section"}}}, \%chap) if ($chap{ "type"} eq "Tips");
        }#for my $chapter
    }
    &report( "done.\n");
}

sub generate_texi {
    move_to_outdir();
    open( OUT, ">$outfile" ) or die $!;
    report( "writing header...", 1);
    make_texi_header();
    report( "done\n");

    report( "writing main menu...", 1);
    make_texi_topmenu();
    report( "done\n");

    report( "writing contents...", 1);
    make_texi_content();
    report( "done\n");

    report( "writing tips menu...", 1);
    make_texi_tips_menu('function index', '');       # パラメータ要確認(前後に位置する index ページをしっかり指定する)
    report( "done\n");
    
    report( "writing closing part...", 1);
    make_texi_closing();
    report( "done\n");

=COMMENT
    report( "checking sections...", 1);
    &check_section( \%INDEX_NAME_LIST, \%sects );
    report( "done\n");
=cut
}

#==============================================================================#

# 状況出力
sub report{
    my( $str, $filename) = map { Jcode::convert($_, "sjis", $in_charset) } @_;
    printf( STDOUT "%s: ", $0) if $filename;
    printf STDOUT $str;
}

# 変換日
sub timestamp{
  sprintf "Converted from %s by %s at %s\n", $datafile, $0, scalar localtime();
}

# 出力ディレクトリへ移動(なければ作る)
sub move_to_outdir {
    unless( -d $outdir){
        mkdir $outdir, 0755;
        &report( "mkdir $outdir\n", 1);
    }
    chdir $outdir or die $!;
}

sub writeln {
    my $strbuf = sprintf(scalar(shift), @_);
    print OUT Jcode::convert($strbuf, $out_charset, $in_charset), "\n";
}

# 一部の文字は加工せにゃならん (`@' と `{' と '}' )
sub encode_text_for_target {
    my $text = shift;
    $text =~ s/(@|{|})/\@$1/g;
    return $text;
}

### [make_texi_header] ##############################
sub make_texi_header {
    my $info_title = "xyzzy Reference";
    my $info_file_description = "xyzzy のリファレンス。";
    my $info_copyright = "(C)1999 ";
    my $infofile;
    ($info_file = $outfile) =~ s@\.[^./\\]+$@@;
    $info_file .= '.info';
    # header
    writeln '\input texinfo   @c -*-texinfo -*-';
    writeln '@setfilename %s', $info_file;
    writeln '@settitle %s', $info_title;
    writeln '@setchapternewpage odd';
    writeln '';
    # file summary
    writeln '@c==[SUMMARY DESCRIPTION AND COPYRIGHT]==============================';
    writeln '@ifinfo';
    writeln '%s', $info_file_description;
    writeln 'Copyright %s', $info_copyright;
    writeln '@end ifinfo';
    writeln '';
    # file title
    writeln '@c==[TITLEPAGE AND COPYRIGHT]==============================';
    writeln '@titlepage';
    writeln '@sp 10';
    writeln '@center @titlefont{%s}', $info_file_description;
    writeln '@page';
    writeln '@vskip 0pt plus 1filll';
    writeln 'Copyright @copyright{} %s', $info_copyright;
    writeln '@end titlepage';
    writeln '';
}

### [make_texi_topmenu] ##############################
sub make_texi_topmenu {
    my $secname;
    local $_;
    # find first section written explicitly.
    for $_ (@INDEX_LIST) {
        if (exists( $sects{ $_ } )) {
            $secname = $_;
            last;
        }
    }

    # top menu beginning
    writeln '@c==[TOP NODE AND MASTER MENU]==============================';
    writeln '@node    Top, %s, (dir), (dir)', get_section_node( $secname );
    writeln '@chapter Top';
    writeln '@menu';

    my $desc_indent = 30;
    my $menufmt = '%1s %-' . $desc_indent . 's%s';
    # sections
    for $secname (@INDEX_LIST) {
        my $doc_exists_mark = exists( $sects{ $secname } ) ? '*' : ' ';
        writeln $menufmt, $doc_exists_mark, get_section_node( $secname ) . '::', $secname;
    }
    # indexes
    writeln '';
    my @indexes = ('tips index::', '目的別索引',
                   'variable index::', '変数索引',
                   'function index::', '関数索引'
                   );
    my $idx = 0;
    while ($idx <= $#indexes) {
        writeln $menufmt, '*', $indexes[$idx++], $indexes[$idx++];
    }
    writeln '@end menu';
    writeln '';
}

### [make_texi_content] ##############################
sub make_texi_content {
    report("\n");
    my (@exist_sects, $section);
    for $section (@INDEX_LIST) {
        push @exist_sects, $section if exists( $sects{ $section } );
    }
    for $idx ( 0..$#exist_sects ) {
        $section = $exist_sects[$idx];
        my @chaps = sort sortby_title (@ { $sects{ $section } });
        # section menu
        make_texi_section_menu(
            $section,
            ( 0 < $idx ) ? $exist_sects[$idx - 1] : "",
            ( $idx < $#exist_sects ) ? $exist_sects[$idx + 1] : "",
            \@chaps);
        # each sections
        for $cidx ( 0..$#chaps ) {
            make_texi_chapter(
                $section,
                $chaps[$cidx],
                ( 0 < $cidx ) ? get_chapter_node( $chaps[$cidx - 1] ) : "",
                ( $cidx < $#chaps ) ? get_chapter_node( $chaps[$cidx + 1] ) : "",
                );
        }
    }
    report("writing all of sections, ");
}
sub sortby_title {
    ( $ { $a } { "title" } ) cmp ( $ { $b } { "title" } );
}
=COMMENT ##############################
sub get_chapter_node {
    my $chap = shift;
    my ($title, $package) = (
        $ { $chap } {"title"},
        $ { $chap } {"package"}
        );
    if ($package) {
        sprintf "%s [%s]", $title, $package;
    } else {
        $title;
    }
}
=cut ##############################
sub get_section_node {
    $ret = $INDEX_NAME_LIST{ scalar(shift) };
    $ret and ($ret .= ' $');
}
sub get_chapter_node {
    my $chap = shift;
    $ { $chap } {"title"};
}
# section menu
sub make_texi_section_menu {
    my ($section, $prev_section, $next_section, $chaps) = @_;
    report("[section] $section\n", 1);
    writeln '@node %s, %s, %s, top',
      get_section_node( $section ),
      get_section_node( $next_section ),
      get_section_node( $prev_section );
    writeln '@chapter %s', $section;
    writeln '@menu';
    for my $chap (@ { $chaps } ) {
        writeln '* %s::', get_chapter_node($chap);
    }
    writeln '@end menu';
    writeln '';
}
# item(defun, defmacro, defvar, etc.)
sub make_texi_chapter {
    my ($section, $chapter, $prev_chap_title, $next_chap_title) = @_;
    my ($title, $args, $type, $package, $desc, $seealso, $file) = (
        $ { $chapter } { "title" },
        $ { $chapter } { "arguments" },
        $ { $chapter } { "type" },
        $ { $chapter } { "package" },
        $ { $chapter } { "description" },
        $ { $chapter } { "seealso" },
        $ { $chapter } { "file" }
        );
    my $cmd = $TEXI_CMD_REF{ $type };
    writeln '@node %s, %s, %s, %s', get_chapter_node($chapter), $next_chap_title, $prev_chap_title, get_section_node( $section );
    # heading
    if ($cmd =~ /^@(.+)$/) {
        # defun, defmacro, defvar and so on.
        my $cmdbody = $1;
        writeln '@chapheading %s', $title;
        writeln '%s %s %s' . "\n", $cmd, $title, $args;
        writeln '@end %s', $cmdbody;
    } else {
        # a chapter about tips/misc
        writeln '@chapheading %s%s', @cmd, $title;
    }
    # property
    if ($package or $file) {
        ($package) and (writeln "[package] %s @*", $package);
        ($file)    and (writeln "[file] %s @*", $file);
        writeln '@*';
    }
    # description
    writeln '@format';
    writeln '%s', $desc;
    writeln '', $desc;
    writeln '@end format';
    # see also
    if ($seealso) {
        map { writeln '@xref{%s} @*', $_ } split "\t", $seealso;
    }
    writeln '';
}

### [make_texi_tips_menu] ##############################
sub make_texi_tips_menu {
    my ($next_index, $prev_index) = @_;
    my (@exist_sects, $section);
    for $section (@INDEX_LIST) {
        push @exist_sects, $section if exists( $tips{ $section } );
    }
    writeln '@c==[CONCEPT INDEX]==============================';
    writeln '@node tips index, %s, %s, top', $next_index, $prev_index;
    writeln '@unnumbered concept index';
    writeln '@menu';
    my $desc_indent = 30;
    my $menufmt = '%1s %-' . $desc_indent . 's%s';
    for $idx ( 0..$#exist_sects ) {
        $section = $exist_sects[$idx];
        my @chaps = sort sortby_title (@ { $tips{ $section } });

        # section separater
        writeln '==[%s]==============================', $section;
        # each sections
        for $cidx ( 0..$#chaps ) {
            writeln '* %s::', get_chapter_node( $chaps[$cidx] );
        }
        writeln '';
    }
    writeln '@end menu';
    writeln '';
}

### [make_texi_closing] ##############################
sub make_texi_closing {
    writeln '@c==[THE END OF THE DOCUMENT]==============================';
    writeln '';
    writeln '@node function index, variable index, tips index, top';
    writeln '@unnumbered function index';
    writeln '@printindex fn';
    
    writeln '@node variable index, , function index, Top';
    writeln '@unnumbered variable index';

    writeln '@printindex vr';

    writeln '@contents';
    writeln '@bye';
}
    
main();
0;